'''
Created on Jul 22, 2016

@author: breno
'''

import os, sys, logging, argparse, subprocess, time
from random import sample
from operator import itemgetter
from shutil import rmtree
from datetime import datetime
try:
    from cPickle import load, dump
except:
    from pickle import load, dump

from coverage_info import get_coverage_info, merge_lcov_files, collect_coverage_info
from selection_algorithms import select_n_greedy_additional


def build_parser():
    """Creates parser for command line arguments """
    parser = argparse.ArgumentParser(description='Test Case Selection')
    parser.add_argument('subject',
                        help='subject name (e.g.: grep, gzip, etc)',
                        type=str)
    parser.add_argument('--ver', '--version',
                        help="subject's version",
                        required=True,
                        dest='version',
                        nargs='?',
                        type=str)
    parser.add_argument('--cleanup',
                        help='Cleanup directory',
                        required=False,
                        dest='cleanup',
                        action='store_true')
    parser.set_defaults(cleanup=False)
    parser.add_argument('-c', '--criterion',
                        help='Adequacy criterion (function, line, or branch)',
                        required=False,
                        dest='entity',
                        nargs='?',
                        type=str)
    parser.add_argument('--fm', '--fault-matrix', 
                        help="fault-matrix types. Either 'all' or 'hard'",
                        required=False,
                        dest='faultmatrix', 
                        nargs='?', 
                        default='all')
    parser.add_argument('-p', '--part',
                        help='Partition ID (e.g.: 1, 2, etc)',
                        required=False,
                        dest='partition',
                        nargs='?',
                        type=int)
    parser.add_argument('-n', '--numtcs',
                        help="Number of test cases for partition 'p'",
                        required=False,
                        dest='num_tcs',
                        nargs='?',
                        type=int)
    parser.add_argument('-i', '--iteration',
                        help="Iteration ID",
                        required=False,
                        dest='iteration',
                        nargs='?',
                        type=int)
    parser.set_defaults(new_iter=False)  
    parser.add_argument('--full-spectrum',
                        help='Considers not covered entities to be part of the importance groups',
                        required=False,
                        dest='full_spectrum',
                        action='store_true')
    parser.set_defaults(full_spectrum=True)
    parser.add_argument('-d','--debug', 
                        help='Print lots of debugging statements', 
                        action="store_const",
                        dest="loglevel",
                        const=logging.DEBUG, 
                        default=logging.WARNING)
    parser.add_argument('-v','--verbose', 
                        help='Be verbose', 
                        action="store_const",
                        dest="loglevel",
                        const=logging.INFO)
    return parser
        

def get_cumulative_spectrum(selected, partition, iteration):
    lcov_files_list = [os.path.join(version_cov_dir, '{}.info'.format(tc)) for tc in selected]        
    #update below
    #info_previous_tcs = os.path.join(spectra_dir, 'spectrum_part{}_iter{}.info'.format(partition, iteration-1))
    
    if iteration > 0 and os.path.exists(spectra_dir):
        last_spectrum_file_name = get_last_spectrum_file(partition, iteration)
        last_spectrum_file = os.path.join(spectra_dir, last_spectrum_file_name)
        #if os.path.exists(info_previous_tcs):
            #logging.debug("File exists: {}".format(info_previous_tcs))
            #lcov_files_list.append(info_previous_tcs)
        if os.path.exists(last_spectrum_file):
            logging.debug("File exists: {}".format(last_spectrum_file))
            lcov_files_list.append(last_spectrum_file)
    
    logging.debug("lcov_files_list: {}".format(lcov_files_list))
    outfile = os.path.join(spectra_dir, 'spectrum_part{}_iter{}.info'.format(partition, iteration))
    if not os.path.exists(spectra_dir):
        os.makedirs(spectra_dir)
    merge_lcov_files(lcov_files_list, outfile)
    #cov_info = get_coverage_info(outfile, format_=['covered', 'reachable'])


def cleanup():
    logging.debug("Cleaning up output from previous runs...")
    if os.path.isfile(already_selected_file):
        os.remove(already_selected_file)
    if os.path.exists(spectra_dir):
        rmtree(spectra_dir)
    os.makedirs(spectra_dir)


def get_entities_imp(count_spectrum, proportions):
    #===========================================================================
    # proportions is a tuple where the first element represents the percentage
    # of entities to be considered "high", the second element represents the 
    # percentage of entities to be considered "medium", and the third element 
    # represents the percentage of entities to be considered "low". An example
    # could be (0.1, 0.8, 0.1). 
    #===========================================================================
    imp = {'high': [], 'medium': [], 'low': []}
    aux_sorted = sorted(count_spectrum, key=itemgetter(1), reverse=True)    
    h = int(round(len(aux_sorted) * proportions[0], 0))
    m = int(round(len(aux_sorted) * proportions[1], 0))
    l = int(round(len(aux_sorted) * proportions[2], 0))
    imp['high'] = [t[0] for t in aux_sorted[:h]]
    imp['medium'] = [t[0] for t in aux_sorted[h:h+m]]
    imp['low'] = [t[0] for t in aux_sorted[h+m:]]
    return imp


def get_revealed_faults(aTestSet, aFaultMatrix):
    revealed = []
    for tc in aTestSet:
        try:
            for fault in aFaultMatrix[tc]:
                if fault not in revealed:
                    revealed.append(fault)
        except KeyError:
            pass
    return sorted(revealed)

def get_last_spectrum_file(partition, iteration):
    assert iteration >= 0, "Something went wrong! Could not find spectrum file!"
    last = 'spectrum_part{}_iter{}.info'.format(partition, iteration)
    while last not in os.listdir(spectra_dir):
            last = get_last_spectrum_file(partition, iteration - 1)
    return last


def main(entity, partition, num_tcs, iteration):
    # At the first iteration, we do not have any spectrum yet (using the online spectrum strategy) 
    # so the test cases are randomly selected from the the target partition
    selected = []
    if iteration == 0:
        #===========================================================================
        # As per Roberto's request, if the number of tcs to be selected is bigger
        # than the number of tcs available in the pool, all the test cases currently
        # available in the pool are returned (prioritized) instead of raising an 
        # exception. The next line has been commented.
        #===========================================================================
        #assert num_tcs <= len(part_map[partition]), "Error! Should select {} test cases but the pool contains {}".format(num_tcs, len(part_map[partition]))
        if num_tcs > len(part_map[partition]):
            logging.info("ATTENTION: want to select {} test cases but the pool contains {}".format(num_tcs, len(part_map[partition])))
            num_tcs = len(part_map[partition])
            logging.info("new numtcs: {}".format(num_tcs))        
        selected = sample(part_map[partition], num_tcs)
        #===========================================================================
    elif iteration >= 1:
        not_selected = [tc for tc in part_map[partition] if tc not in already_selected]        
        spectrum_file_name = get_last_spectrum_file(partition, iteration)
        #spectrum_file = os.path.join(spectra_dir, 'spectrum_part{}_iter{}.info'.format(partition, iteration-1))
        spectrum_file = os.path.join(spectra_dir, spectrum_file_name)
        if args.full_spectrum:
            count_spectrum_all = get_coverage_info(spectrum_file, format_=['detailed'])
            count_spectrum = count_spectrum_all['{}_{}'.format('detailed', entity)]
        else:
            count_spectrum_all = get_coverage_info(spectrum_file, format_=['coveredcount'])
            count_spectrum = count_spectrum_all['{}_{}'.format('coveredcount', entity)]
        importance_groups = get_entities_imp(count_spectrum, hml_proportion)
        #expected_cov_op = [ent for vlist in importance_groups.values() for ent in vlist]  # rework on this later
        expected_cov_op = [ent[0] for ent in count_spectrum]
        selected = select_n_greedy_additional(num_tcs, not_selected, expected_cov_op, coverage_info, mode='weighted', entities_importance=importance_groups, weights=hml_weights)
    
    # Updated spectrum
    get_cumulative_spectrum(selected, partition, iteration)
    logging.info("Selected {} test cases for partition {}: {}".format(len(selected), partition, selected))
    selected_str = ",".join(str(i) for i in selected)
    print selected_str

    already_revealed = get_revealed_faults(already_selected, fault_matrix)
    revealed_by_selected = get_revealed_faults(selected, fault_matrix)
    new_revealed = [fault for fault in revealed_by_selected if fault not in already_revealed]
    logging.info("Faults revealed on previous iterations: {}".format(already_revealed))
    logging.info("[ALL] faults revealed at this iteration: {}".format(revealed_by_selected))
    logging.info("[NEW] faults revealed at this iteration: {}".format(new_revealed))
        
    # Updated "already selected list"
    already_selected.extend(selected)
    with open(already_selected_file, "w") as _:
        dump(already_selected, _)
    
    #print len(new_revealed)
    logging.info("# of faults revealed: {}".format(len(new_revealed)))
    #return len(new_revealed)


if __name__ == '__main__':
    """ Main entry point. Reads the command line arguments and performs the requested actions"""    
    # Parse the command line arguments
    argument_parser = build_parser()
    args = argument_parser.parse_args()
    
    #===========================================================================
    # # Logging options
    # # create file handler which logs even debug messages
    # fh = logging.FileHandler('python_log.log')
    # fh.setLevel(logging.DEBUG)
    # # create formatter and add it to the handlers
    # formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    # fh.setFormatter(formatter)
    # # add the handlers to the logger
    # logger.addHandler(fh)
    #===========================================================================

    
    # Set logging level
    #logging.basicConfig(level=args.loglevel)
    logging.basicConfig(level=args.loglevel,
                        format='%(asctime)s %(name)-12s %(levelname)-8s %(message)s',
                        datefmt='%m-%d %H:%M',
                        #filename='python_log.log',
                        filemode='a')
    #logging.info("{}".format("="*80))
    #logging.info("{}".format(str(datetime.now())))
    #logging.info("{}".format("="*80))

    script_loc = os.path.dirname(os.path.realpath(__file__))
    subject_dir = os.path.join(script_loc, 'subjects', args.subject)
    version_dir = os.path.join(subject_dir, args.version) 
    version_cov_dir = os.path.join(version_dir, 'individual_test_coverage')   
    spectra_dir = os.path.join(version_dir, 'spectrum_files')   
    already_selected_file = os.path.join(version_dir, "already_selected.pickle")
    
    #=========================================================================== 
    # PRE_SETUP
    #===========================================================================
    # Cleaning up (if needed)
    if args.cleanup:
        cleanup()
        sys.exit()
    #if args.iteration == 0 and args.partition == 0:  # do a clenaupfirst
        #cleanup()
    
    # Loading fault-matrix
    if args.faultmatrix == 'all':
        fm_file = 'fault_matrix'
        logging.info('default fault-matrix will be used')
    elif args.faultmatrix == 'hard':
        fm_file = 'fault_matrix_hard'
        logging.info('fault-matrix with "hard-to-find faults" will be used')
    with open(os.path.join(version_dir, 'fault_matrices', '{}_key_tc.pickle'.format(fm_file)), 'r') as file_t:
        fault_matrix = load(file_t)
    
    #===========================================================================
    # Single dictionary
    #===========================================================================
    # Creating Coverage info dictionary if it does not exist
    #===========================================================================
    # if not os.path.exists(os.path.join(version_dir, 'coverage_info.pickle')):
    #     collect_coverage_info(version_dir, format_=[cov_info_format], dump_data=True)
    # with open(os.path.join(version_dir, 'coverage_info.pickle'), 'r') as file_:
    #     coverage_info_aux = load(file_)
    #     coverage_info = {}
    #     for tc in coverage_info_aux:
    #         coverage_info[tc] = coverage_info_aux[tc]['{}_{}'.format(cov_info_format, args.entity)]
    #===========================================================================
    
    #===========================================================================
    # One dictionary per criterion
    #===========================================================================    
    # Creating Coverage info dictionary if it does not exist
    if not os.path.exists(os.path.join(version_dir, 'coverage_info_{}.pickle'.format(args.entity))):
        collect_coverage_info(version_dir, format_=['coveredcount'], dump_data=True, split_dict=True)
    # Loading Coverage info dictionary
    with open(os.path.join(version_dir, 'coverage_info_{}.pickle'.format(args.entity)), 'r') as file_:
        coverage_info = load(file_)        
    #===========================================================================
    
    # Loading partition mapping dictionary    
    with open(os.path.join(version_dir, 'part_map.pickle'), 'r') as file_:
        part_map = load(file_)
    
    # Loading list of "already selected" test cases
    if os.path.exists(already_selected_file):
        with open(already_selected_file, 'r') as file_:
            already_selected = load(file_)
    else:
        already_selected = []
    
    # Getting list of test cases for target subject
    tc_list = coverage_info.keys()
    #===========================================================================
    #logging.info('Processing {} version {}'.format(args.subject, args.version))
    logging.info("="*80)
    logging.info("ITERATION #{} - PARTITION #{}".format(args.iteration, args.partition))
    logging.info("Total # of tcs executed on previous iterations: {}".format(len(already_selected)))
    #===========================================================================
    
    hml_proportion = (1/3.0, 1/3.0, 1/3.0)
    hml_weights = {'high': 0.1, 'medium': 1.0, 'low': 10.0}
    main(args.entity, args.partition, args.num_tcs, args.iteration)

    
    
